/* Resolution Method for Coalition Logic     */
/* Delivers all proofs                       */
/* No strategies                             */
/* Remove repeated clauses                   */
/* Remove from I clauses repeated in U       */
/* Propositional forward subsumption         */
/* Get rid of tautologies                    */
/*                                           */
/* Cludia Nalon, Liverpool, 27/02/2013      */

/* Problems are given in the following form:   

   [ [Initial Clauses],
     [Global Clauses],
     [Positive Coalitional Clauses],
     [Negative Coalitional Clauses]]          */

/* Initial and global clauses are list of literals */

/* A coalitional clause is in the following form: 

   [ [Terms],
     [Agents],
     [Clause] ]

where Terms and Clause are both lists of literals, 
and Agents is a list. 
Note that Agents \subseteq AllAgents is not checked.

*/

?-[definitions,rules,subsumption,output].
?-dynamic clprover_settings/2.

prove:-current_prolog_flag(argv, Arguments),
       Arguments = [_SystemArg|Args],
       %%% defaults: read formula from command line and outputs the result to screen
       assertz(clprover_settings(mode_input,standard)),
       assertz(clprover_settings(name_input,'')),
       assertz(clprover_settings(formula,[[],[],[],[]])),
       assertz(clprover_settings(mode_output,standard)),
       assertz(clprover_settings(name_output,screen)),
       treat_args(Args),
       clprover_settings(mode_input,MI),
       clprover_settings(name_input,NI),
       clprover_settings(mode_output,MO),
       clprover_settings(name_output,NO),
       prove(MI,NI,MO,NO),nl,
       halt.
       
%       append(_SytemArgs, [--|Args], Arguments), !,
%       Args=[Input],
%       atom_to_term(Input,Problem,_),
%       prove(Problem),halt.

treat_args([]).
treat_args(['-i',InputName|Args]):-
       retractall(clprover_settings(mode_input,_)),
       retractall(clprover_settings(name_input,_)),
       assertz(clprover_settings(mode_input,file)),
       assertz(clprover_settings(name_input,InputName)),
       treat_args(Args).
treat_args(['-f',Formula|Args]):-
       retractall(clprover_settings(mode_input,_)),
       retractall(clprover_settings(formula,_)),
       assertz(clprover_settings(mode_input,standard)),
       atom_to_term(Formula,Problem,_),
       assertz(clprover_settings(formula,Problem)),
       treat_args(Args).
treat_args(['-o',OutputName|Args]):-
       retractall(clprover_settings(mode_output,_)),
       retractall(clprover_settings(name_output,_)),
       assertz(clprover_settings(mode_output,latex)),
       assertz(clprover_settings(name_output,OutputName)),
       treat_args(Args).
treat_args(['-h'|_]):-
       nl,write('Usage help: '),nl,
       nl,write('-i <file> :: input problems are read from <file>'),
       nl,write('-o <file> :: proofs are written in latex format in <file>'),
       nl,write('-f <problem> :: problem is read from command line'),nl,
       halt.
treat_args(_):-
      treat_args(['-h'|_]).

prove(standard,_,standard,_):-
       clprover_settings(formula,Problem),
       time(resolution_proof(Problem,Derivation)),!,%,
       ((member([_,[[],[],[]],_,_],Derivation),write('Unsatisfiable'),nl)|(write('Satisfiable'),nl)).

prove(standard,_,latex,Output):-
       clprover_settings(formula,Problem),
       time(resolution_proof(Problem,Derivation)),!,%,
       sort(Derivation,SDerivation),
       open(Output,write,OFile),
       write(OFile,'%% Generated by cl_prover.pl'),
       nl(OFile),
       write(OFile,'\\begin{longtable}{rrcll}'),nl(OFile),
       write_derivation(OFile,SDerivation),
       write(OFile,'\\end{longtable}'),nl(OFile),
       nl(OFile),
       ((member([_,[[],[],[]],_,_],Derivation),write('Unsatisfiable'),nl,nl(OFile),write(OFile,'Unsatisfiable'))|(write('Satisfiable'),nl,nl(OFile),write(OFile,'Satisfiable'))),
       close(OFile).

prove(file,Input,standard,screen):-
   open(Input,read,IFile),
   write_derivations(IFile,screen),
   close(IFile).

prove(file,Input,latex,Output):-
   open(Input,read,IFile),
   open(Output,write,OFile),
   write(OFile,'%% Generated by cl_prover.pl'),
   nl(OFile),
   write_derivations(IFile,OFile),
   close(OFile),
   close(IFile).
       
write_derivations(IFile,_):-
   at_end_of_stream(IFile).

write_derivations(IFile,screen):-
       read_term(IFile,Problem,[comments(_Comment)]),
garbage_collect,trim_stacks,
       time(resolution_proof(Problem,Derivation)),!,%,
       ((member([_,[[],[],[]],_,_],Derivation),write('Unsatisfiable'),nl)|(write('Satisfiable'),nl)),
       write_derivations(IFile,screen).

write_derivations(IFile,OFile):-
       read_term(IFile,Problem,[comments(_Comment)]),
garbage_collect,trim_stacks,
       time(resolution_proof(Problem,Derivation)),!,%,
       ((member([_,[[],[],[]],_,_],Derivation),write('Unsatisfiable'),nl,nl(OFile),write(OFile,'Unsatisfiable'))|(write('Satisfiable'),nl,nl(OFile),write(OFile,'Satisfiable'))),
       sort(Derivation,SDerivation),
       nl(OFile),
       write(OFile,'\\begin{longtable}{rrcll}'),nl(OFile),
       write_derivation(OFile,SDerivation),
       write(OFile,'\\end{longtable}'),nl(OFile),
       write(OFile,'\\pagebreak'),nl(OFile),
       write_derivations(IFile,OFile).

resolution(Output,Problem):-
   resolution_proof(Problem,Derivation),
   ((member([_,[[],[],[]],_,_],Derivation),write('Unsatisfiable'),nl)|(write('Satisfiable'),nl)),%,
   sort(Derivation,SDerivation),
   write_proof(Output,SDerivation).

resolution_proof(Clauses,Derivation):-
   clear,
   number_clauses_in_problem(Clauses,NumberedClauses),
   resolve_problem(NumberedClauses,Derivation).

number_clauses_in_problem([Initial,Global,Positive,Negative],
                          [NInitial,NGlobal,NPositive,NNegative]):-
   number_clauses(Initial,NInitial,i),
   number_clauses(Global,NGlobal,u),
   number_clauses(Positive,NPositive,p),
   number_clauses(Negative,NNegative,n).
   
number_clauses([],[],_).
number_clauses([Head|Tail],[NewHead|NewTail],Set):-
   newclausecount(N),
   sort_head(Head,Set,SHead),
   NewHead = [N,SHead,[given],Set],
   number_clauses(Tail,NewTail,Set).

sort_head(Head,Set,SHead):-
   ((Set == i | Set == u)),
   sort(Head,SHead).
sort_head([LHS,A,C],_Set,[SLHS,SA,SC]):-
   sort(LHS,SLHS),
   sort(A,SA),
   sort(C,SC).

resolve_problem([Initial,Global,Positive,Negative],Derivation):-
   modal(Global,MGlobal),
%   nl,write('Global Resolution:'),nl,
garbage_collect,trim_stacks,
   remove_self_subsumed(MGlobal,CMGlobal),
   subtract(CMGlobal,[[]],FCMGlobal),
   cycle(FCMGlobal,[],GDerived),
   append(Positive,Negative,Coalitional),
%   nl,write('Global and Coalitional Resolution:'),nl,
garbage_collect,trim_stacks,
   cycle(Coalitional,GDerived,Derived),
   findall([N,[[],[],C],J,u],member([N,[[],[],C],J,u],Derived),AllGClauses),
   ((member([_,[[],[],[]],_,_],AllGClauses),Derivation=Derived) |
    (subtract(Derived,AllGClauses,CDerived),
     modal(Initial,MInitial),
%   nl,write('Initial Resolution:'),nl,
     garbage_collect,trim_stacks,
     remove_self_subsumed(MInitial,CMInitial),
     subtract(CMInitial,[[]],FCMInitial),
     remove_subsumed(FCMInitial,Derived,NewInitial),
     cycle(NewInitial,AllGClauses,IDerivation),
     append(IDerivation,CDerived,Derivation))).

cycle([],AllClauses,AllClauses).%:-nl,write('No new clauses generated (1).'),nl.

cycle(NewClauses,AllClauses,Derivation):-
%   write('Cycle - New Clauses are: '),nl,%trace,
%   write_proof(screen,NewClauses),nl,nl,%read(P),
garbage_collect,trim_stacks,
   resolve_clauses(NewClauses,AllClauses,Derived),
garbage_collect,trim_stacks,
   flatten_derived(Derived,FDerived),%nl,
   remove_repeated(FDerived,NDerived),
   append(NewClauses,AllClauses,Clauses),
   subtract(NDerived,Clauses,NewC1),
garbage_collect,trim_stacks,
   remove_repeated(NewC1,Clauses,NewC2),
%   nl,write('Generated Clauses are:'),nl,write_proof(screen,NewC2),nl,
   generate_globals(NewC2,Globals),
%   nl,write('Globals from Rewriting'),nl,write_proof(screen,Globals),nl,
   remove_repeated(Globals,NewGlobals),
   remove_repeated(NewGlobals,Clauses,NewestGlobals),
   append(NewC2,NewestGlobals,NewAll),
garbage_collect,trim_stacks,
   remove_self_subsumed(NewAll,NewerAll), 
   subtract(NewerAll,[[]],FNewerAll),
   remove_subsumed(FNewerAll,Clauses,NewestAll),
   check(NewestAll,Clauses,Derivation).

flatten_derived([],[]).
flatten_derived([H|T],H):-
   subset(T,[]).
flatten_derived([H1,H2|Tail],NewTail):-
   append(H1,H2,H),
   flatten_derived([H|Tail],NewTail).


check([],AllClauses,AllClauses).%:-nl,write('No new clauses generated (2).'),nl.

check(C1,C2,Derivation):-
   cycle(C1,C2,Derivation).

